//   
// VECT_TAB_SRAM    RAM!

#include <stm32f10x.h>

//  DIGS_CONTROL, SEGS_CONTROL 
//     .
//   - ;   
//   -  (0 - GPIOA,  ..);
//      ,  
//   .

//     
//   .   :
// DIG1 - PC6, DIG2 - PC7, DIG3 - PC8, DIG4 - PC9.
const char DIGS_CONTROL[]={0x26, 0x27, 0x28, 0x29};

const uint32_t DIGS_COUNT=
		sizeof DIGS_CONTROL/sizeof DIGS_CONTROL[0];

//     
//  .   :
// A - PA8, B - PA9, C - PA10, D - PA11,
// E - PA12, F - PC0, G - PC1, DP - PC2.
const char SEGS_CONTROL[]={0x08, 0x09, 0x0A, 0x0B,
		0x0C, 0x20, 0x21, 0x22};

//   -  .
GPIO_TypeDef* const PORTS[]={GPIOA, GPIOB,
			GPIOC, GPIOD};

//    -, 
//  (   XXXX_CONTROL)   state.
void line_set(char port_pin, bool state)
{
	if(state)
		PORTS[(port_pin>>4)&0xF]->BSRR=1<<(port_pin&0xF);
	else
		PORTS[(port_pin>>4)&0xF]->BRR=1<<(port_pin&0xF);

}
//   (.. )  
//  ; . 0 (false) -  
//   ; . 1 (true) -     .
//      
//     .
#define DIG_ACTIVE_LEVEL false

//  ()   ;
// . 1     ;
// . 0     .
//     
//   ,  
//   .
#define SEG_ACTIVE_LEVEL true

//    .
void digs_off()
{
	for(uint32_t i=0; i<DIGS_COUNT; i++)
		line_set(DIGS_CONTROL[i], !DIG_ACTIVE_LEVEL);
}

//   d,    0;
// d=0    .
void dig_on(uint32_t d)
{
	line_set(DIGS_CONTROL[d], DIG_ACTIVE_LEVEL);
}

//       b;
//     A,  - B
//  ..,      DP.
//    ,  .
void set_seg_bus(char b)
{
	for(uint32_t i=0; i<8; i++, b>>=1)
		line_set(SEGS_CONTROL[i], (b&1)^!SEG_ACTIVE_LEVEL);
}

//   ,
//  .
void pin_config(char pp, bool level)
{
	uint32_t port=(pp>>4)&0xF;
	uint32_t pin=pp&0xF;
	//    
	// (,    
	//  ).
	RCC->APB2ENR|=RCC_APB2ENR_IOPAEN<<port;
	if(level)
		PORTS[port]->BSRR=1<<pin;
	else
		PORTS[port]->BRR=1<<pin;
	uint32_t mask, fnmode;
	volatile uint32_t *mode_reg=(pin<8)?&PORTS[port]->CRL:
			&PORTS[port]->CRH;
	if(pin>=8)
		pin-=8;
	mask=~(0xF<<(4*pin));
	fnmode=2<<(4*pin);
	*mode_reg&=mask;
	*mode_reg|=fnmode;
}

//   , 
//   .
void display_config()
{
	for(uint32_t i=0; i<DIGS_COUNT; i++)
		pin_config(DIGS_CONTROL[i], !DIG_ACTIVE_LEVEL);
	for(uint32_t i=0; i<8; i++)
			pin_config(SEGS_CONTROL[i], !SEG_ACTIVE_LEVEL);
}

//  ,   
//      .
//   ASCII  !
//       
// 0..9,    '0'..'9';
// 10..15   'A', 'b', 'C', 'd', 'E', 'F';
// 16 -  (  );
// 17 - '-'.
//  ,    
//  .
const char CG_TABLE[]=
	{0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, // '0'..'5'
	 0x7D, 0x07, 0x7F, 0x6F,			 // '6'..'9'
	 0x77, 0x7C, 0x39, 0x5E, 0x79, 0x71, // 'A'..'F'
	 0x00, // SPACE
	 0x40  // MINUS
	};
#define SPACE 16
#define MINUS 17

//      .
//    ASCII ,  
//    
// ( 0 -  '0'  ..).
//       
//       .
//        .
char OUT_STR[]={SPACE, SPACE, SPACE, SPACE};

//      ;
//      f=N*F,
//  N -   ,
// F -    .
void display_refresh()
{
	const uint32_t str_len=
			sizeof OUT_STR/sizeof OUT_STR[0];
	static uint32_t d=0;
	char ch=OUT_STR[d];
	char dot=ch&0x80;
	ch&=0x7F;
	ch=CG_TABLE[ch]|dot;
	digs_off();
	set_seg_bus(ch);
	dig_on(d);
	if(++d>=str_len)
		d=0;
}

//  .
volatile uint32_t ms_counter=0;

//     .
extern "C" void SysTick_Handler()
{
	//     f=500 .
	if(ms_counter&1)
		display_refresh();
	ms_counter++;
}

//    
//  ms .
void delay(uint32_t ms)
{
	uint32_t t0=ms_counter;
	while(ms_counter-t0<ms);
}

//      .
//     ,
//    : 'E'  '-E'
//    false.
bool display(int n)
{
	char sign=SPACE;
	if(n<0)
	{
		n=-n;
		sign=MINUS;
	}
	char d=0;
	for(uint32_t i=0; i<DIGS_COUNT; i++)
	{
		if(n!=0)
		{
			d=n%10;
			n/=10;
		}
		if(i==0)
			d|=0x80; //    .
		else if(n==0&&d==0)
		{
			//  ,  ;
			//   .
			d=sign;
			sign=SPACE;
		}
		OUT_STR[DIGS_COUNT-1-i]=d;
		d=0;
	}
	if(n!=0||sign!=SPACE)
	{
		for(uint32_t i=0; i<DIGS_COUNT-1; i++)
			OUT_STR[i]=SPACE;
		OUT_STR[DIGS_COUNT-1]=0xE;
		if(sign==MINUS&&DIGS_COUNT>=2)
			OUT_STR[DIGS_COUNT-2]=MINUS;
		return false;
	}
	return true;
}

int main()
{
	//   ,
	//  .
	display_config();
	//   .
	SysTick_Config(SystemCoreClock/1000);

	//        .
	OUT_STR[0]=0x88;
	OUT_STR[1]=0x88;
	OUT_STR[2]=0x88;
	OUT_STR[3]=0x88;
	delay(1500);

	//  .
	OUT_STR[0]=SPACE;
	OUT_STR[1]=SPACE;
	OUT_STR[2]=SPACE;
	OUT_STR[3]=SPACE;
	delay(500);

	//   
	//    (   ).
	const uint32_t N=sizeof CG_TABLE/sizeof CG_TABLE[0];
	for(uint32_t i=0; i<2*N; i++)
	{
		char c=i/2;
		if(i&1) c|=0x80;
		OUT_STR[DIGS_COUNT-1]=c;
		delay(1000);
	}

	//  4- .
	display(-1000);
	delay(2000);
	display(10000);
	delay(2000);

	//      -123.
	uint32_t t=-123;
	while(true)
	{
		display(t++);
		delay(1000);
	}
}
